@@ -2,12 +2,12 @@ package App::perlbrew;
 use strict;
 use warnings;
 use 5.008;
-our $VERSION = "0.63";
+our $VERSION = "0.64";
 
 BEGIN {
     ### Special treat for Cwd to prevent it to be loaded from a local::lib dir that is not binary-compatible with system perl.
     if (my $perl5lib = $ENV{PERL5LIB}) {
-        local $ENV{PERL5LIB} = undef;
+        delete $ENV{PERL5LIB};
         require Cwd;
         $ENV{PERL5LIB} = $perl5lib;
     }
@@ -28,6 +28,45 @@ our $CONFIG;
 our $PERLBREW_ROOT = $ENV{PERLBREW_ROOT} || joinpath($ENV{HOME}, "perl5", "perlbrew");
 our $PERLBREW_HOME = $ENV{PERLBREW_HOME} || joinpath($ENV{HOME}, ".perlbrew");
 
+my @flavors = ( { d_option => 'usethreads',
+                  implies  => 'multi',
+                  common   => 1,
+                  opt      => 'thread|threads' }, # threads is for backward compatibility
+
+                { d_option => 'usemultiplicity',
+                  opt      => 'multi' },
+
+                { d_option => 'uselongdouble',
+                  common   => 1,
+                  opt      => 'ld' },
+
+                { d_option => 'use64bitint',
+                  common   => 1,
+                  opt      => '64int' },
+
+                { d_option => 'use64bitall',
+                  implies  => '64int',
+                  opt      => '64all' },
+
+                { d_option => 'DEBUGGING',
+                  opt      => 'debug' }
+              );
+
+
+my %flavor;
+my $flavor_ix = 0;
+for (@flavors) {
+    my ($name) = $_->{opt} =~ /([^|]+)/;
+    $_->{name} = $name;
+    $_->{ix} = ++$flavor_ix;
+    $flavor{$name} = $_;
+}
+for (@flavors) {
+    if (my $implies = $_->{implies}) {
+        $flavor{$implies}{implied_by} = $_->{name};
+    }
+}
+
 ### functions
 
 sub joinpath { join "/", @_ }
@@ -39,7 +78,7 @@ sub mkpath {
 
 sub rmpath {
     require File::Path;
-    File::Path::rmtree([@_], 0, 1);
+    File::Path::rmtree([@_], 0, 0);
 }
 
 sub min(@) {
@@ -81,9 +120,9 @@ sub files_are_the_same {
         if (! @command) {
             my @commands = (
                 # curl's --fail option makes the exit code meaningful
-                [qw( curl --silent --location --fail --insecure )],
+                [qw( curl --silent --location --fail )],
+                [qw( wget --quiet -O - )],
                 [qw( fetch -o - )],
-                [qw( wget --no-check-certificate --quiet -O - )],
             );
             for my $command (@commands) {
                 my $program = $command->[0];
@@ -120,6 +159,7 @@ sub files_are_the_same {
 sub perl_version_to_integer {
     my $version = shift;
     my @v = split(/[\.\-_]/, $version);
+    return undef if @v < 2;
     if ($v[1] <= 5) {
         $v[2] ||= 0;
         $v[3] = 0;
@@ -166,8 +206,13 @@ sub new {
         U => [],
         A => [],
         sitecustomize => '',
+        noman => '',
+        variation => '',
+        both => [],
     );
 
+    $opt{$_} = '' for keys %flavor;
+
     # build a local @ARGV to allow us to use an older
     # Getopt::Long API in case we are building on an older system
     local (@ARGV) = @argv;
@@ -196,6 +241,8 @@ sub new {
 sub parse_cmdline {
     my ($self, $params, @ext) = @_;
 
+    my @f = map { $flavor{$_}{opt} || $_ } keys %flavor;
+
     Getopt::Long::GetOptions(
         $params,
 
@@ -208,6 +255,7 @@ sub parse_cmdline {
         'version',
         'root=s',
         'switch',
+        'all',
 
         # options passed directly to Configure
         'D=s@',
@@ -217,6 +265,13 @@ sub parse_cmdline {
         'j=i',
         # options that affect Configure and customize post-build
         'sitecustomize=s',
+        'noman',
+
+        # flavors support
+        'both|b=s@',
+        'all-variations',
+        'common-variations',
+        @f,
 
         @ext
     )
@@ -290,7 +345,7 @@ sub configure_args {
 sub cpan_mirror {
     my ($self, $v) = @_;
     unless($self->{cpan_mirror}) {
-        $self->{cpan_mirror} = $self->env("PERLBREW_CPAN_MIRROR") || "http://search.cpan.org/CPAN";
+        $self->{cpan_mirror} = $self->env("PERLBREW_CPAN_MIRROR") || "http://www.cpan.org";
         $self->{cpan_mirror} =~ s{/+$}{};
     }
     return $self->{cpan_mirror};
@@ -437,12 +492,6 @@ sub run_command {
         }
     }
 
-    if ($x eq 'install') {
-        # prepend "perl-" to version number, but only if there is an argument
-        $args[0] =~ s/\A((?:\d+\.)*\d+)\Z/perl-$1/
-            if @args;
-    }
-
     $self->$s(@args);
 }
 
@@ -572,7 +621,9 @@ sub run_command_available {
 sub available_perls {
     my ( $self, $dist, $opts ) = @_;
 
-    my $url = "http://www.cpan.org/src/README.html";
+    my $url = $self->{all}  ? "http://www.cpan.org/src/5.0/"
+                            : "http://www.cpan.org/src/README.html" ;
+
     my $html = http_get( $url, undef, undef );
 
     unless($html) {
@@ -582,8 +633,14 @@ sub available_perls {
     my @available_versions;
 
     for ( split "\n", $html ) {
-        push @available_versions, $1
-          if m|<td><a href="http://www.cpan.org/src/.+?">(.+?)</a></td>|;
+        if ( $self->{all} ) {
+            push @available_versions, $1
+                if m|<a href="perl.*?\.tar\.gz">(.+?)</a>|;
+        }
+        else {
+            push @available_versions, $1
+                if m|<td><a href="http://www.cpan.org/src/.+?">(.+?)</a></td>|;
+        }
     }
     s/\.tar\.gz// for @available_versions;
 
@@ -593,6 +650,20 @@ sub available_perls {
 sub perl_release {
     my ($self, $version) = @_;
 
+    # try src/5.0 symlinks, either perl-5.X or perl5.X; favor .tar.bz2 over .tar.gz
+    my $index = http_get("http://www.cpan.org/src/5.0/");
+    if ($index) {
+        for my $prefix ( "perl-", "perl" ){
+            for my $suffix ( ".tar.bz2", ".tar.gz" ) {
+                my $dist_tarball = "$prefix$version$suffix";
+                my $dist_tarball_url = $self->cpan_mirror() . "/src/5.0/$dist_tarball";
+                return ( $dist_tarball, $dist_tarball_url )
+                    if ( $index =~ /href\s*=\s*"\Q$dist_tarball\E"/ms );
+            }
+        }
+    }
+
+    # try CPAN::Perl::Releases
     require CPAN::Perl::Releases;
     my $tarballs = CPAN::Perl::Releases::perl_tarballs($version);
 
@@ -604,12 +675,13 @@ sub perl_release {
         return ($dist_tarball, $dist_tarball_url);
     }
 
+    # try to find it on search.cpan.org
     my $mirror = $self->config->{mirror};
     my $header = $mirror ? { 'Cookie' => "cpan=$mirror->{url}" } : undef;
     my $html = http_get("http://search.cpan.org/dist/perl-${version}", $header);
 
     unless ($html) {
-        die "ERROR: Failed to download perl-${version} tarball.";
+        die "ERROR: Failed to locate perl-${version} tarball.";
     }
 
     my ($dist_path, $dist_tarball) =
@@ -799,15 +871,26 @@ sub do_extract_tarball {
     my $self = shift;
     my $dist_tarball = shift;
 
+    # Assuming the dir extracted from the tarball is named after the tarball.
+    my $dist_tarball_basename = $dist_tarball;
+    $dist_tarball_basename =~ s{.*/([^/]+)\.tar\.(?:gz|bz2)$}{$1};
+
+    # Note that this is incorrect for blead.
+    my $extracted_dir = "@{[ $self->root ]}/build/$dist_tarball_basename";
+
     # Was broken on Solaris, where GNU tar is probably
     # installed as 'gtar' - RT #61042
     my $tarx =
         ($^O eq 'solaris' ? 'gtar ' : 'tar ') .
         ( $dist_tarball =~ m/bz2$/ ? 'xjf' : 'xzf' );
+
+    if (-d $extracted_dir) {
+        rmpath($extracted_dir);
+    }
+
     my $extract_command = "cd @{[ $self->root ]}/build; $tarx $dist_tarball";
     die "Failed to extract $dist_tarball" if system($extract_command);
-    $dist_tarball =~ s{.*/([^/]+)\.tar\.(?:gz|bz2)$}{$1};
-    return "@{[ $self->root ]}/build/$dist_tarball"; # Note that this is incorrect for blead
+    return $extracted_dir;
 }
 
 sub do_install_blead {
@@ -852,7 +935,7 @@ sub do_install_blead {
     return;
 }
 
-sub resolve_stable {
+sub resolve_stable_version {
     my ($self) = @_;
 
     my ($latest_ver, $latest_minor);
@@ -867,11 +950,11 @@ sub resolve_stable {
     die "Can't determine latest stable Perl release\n"
         if !defined $latest_ver;
 
-    return "perl-$latest_ver";
+    return $latest_ver;
 }
 
 sub do_install_release {
-    my ($self, $dist, $dist_name, $dist_version) = @_;
+    my ($self, $dist, $dist_version) = @_;
 
     my ($dist_tarball, $dist_tarball_url) = $self->perl_release($dist_version);
     my $dist_tarball_path = joinpath($self->root, "dists", $dist_tarball);
@@ -881,7 +964,7 @@ sub do_install_release {
             if $self->{verbose};
     }
     else {
-        print "Fetching $dist_name $dist_version as $dist_tarball_path\n" unless $self->{quiet};
+        print "Fetching perl $dist_version as $dist_tarball_path\n" unless $self->{quiet};
         $self->download( $dist_tarball_url, $dist_tarball_path );
     }
 
@@ -898,48 +981,147 @@ sub run_command_install {
         exit(-1);
     }
 
-    $dist = $self->resolve_stable if $dist =~ m/^(?:perl-?)?stable$/;
+    $self->{dist_name} = $dist; # for help msg generation, set to non
+                                # normalized name
 
-    $self->{dist_name} = $dist;
+    if ($dist =~ /^(?:perl-?)?([\d._]+(?:-RC\d+)?|git|stable|blead)$/) {
+        my $version = ($1 eq 'stable' ? $self->resolve_stable_version : $1);
+        $dist = "perl-$version"; # normalize dist name
 
-    my $installation_name = $self->{as} || $dist;
-    if ($self->is_installed( $installation_name ) && !$self->{force}) {
-        die "\nABORT: $installation_name is already installed.\n\n";
-    }
-
-    my $help_message = "Unknown installation target \"$dist\", abort.\nPlease see `perlbrew help` for the instruction on using the install command.\n\n";
-
-    my ($dist_name, $dist_version) = $dist =~ m/^(perl)-?([\d._]+(?:-RC\d+)?|git)$/;
-    if (!$dist_name || !$dist_version) { # some kind of special install
-        if (-d "$dist/.git") {
-            $self->do_install_git($dist);
+        my $installation_name = ($self->{as} || $dist) . $self->{variation};
+        if (not $self->{force} and $self->is_installed( $installation_name )) {
+            die "\nABORT: $installation_name is already installed.\n\n";
         }
-        if (-f $dist) {
-            $self->do_install_archive($dist);
-        }
-        elsif ($dist =~ m/^(?:https?|ftp|file)/) { # more protocols needed?
-            $self->do_install_url($dist);
-        }
-        elsif ($dist =~ m/(?:perl-)?blead$/) {
+
+        if ($version eq 'blead') {
             $self->do_install_blead($dist);
         }
         else {
-            die $help_message;
+            $self->do_install_release( $dist, $version );
         }
+
     }
-    elsif ($dist_name eq 'perl') {
-        $self->do_install_release( $dist, $dist_name, $dist_version );
+    # else it is some kind of special install:
+    elsif (-d "$dist/.git") {
+        $self->do_install_git($dist);
+    }
+    elsif (-f $dist) {
+        $self->do_install_archive($dist);
+    }
+    elsif ($dist =~ m/^(?:https?|ftp|file)/) { # more protocols needed?
+        $self->do_install_url($dist);
     }
     else {
-        die $help_message;
+        die "Unknown installation target \"$dist\", abort.\nPlease see `perlbrew help` " .
+            "for the instruction on using the install command.\n\n";
     }
 
-    $self->switch_to($installation_name)
-        if $self->{switch};
-
+    if ($self->{switch}) {
+        if (defined(my $installation_name = $self->{installation_name})) {
+            $self->switch_to($installation_name)
+        }
+        else {
+            warn "can't switch, unable to infer final destination name.\n\n";
+        }
+    }
     return;
 }
 
+sub check_and_calculate_variations {
+    my $self = shift;
+    my @both = @{$self->{both}};
+
+    if ($self->{'all-variations'}) {
+        @both = keys %flavor;
+    }
+    elsif ($self->{'common-variations'}) {
+        push @both, grep $flavor{$_}{common}, keys %flavor;
+    }
+
+    # check the validity of the varitions given via 'both'
+    for my $both (@both) {
+        $flavor{$both} or die "$both is not a supported flavor.\n\n";
+        $self->{$both} and die "options --both $both and --$both can not be used together";
+        if (my $implied_by = $flavor{$both}{implied_by}) {
+            $self->{$implied_by} and die "options --both $both and --$implied_by can not be used together";
+        }
+    }
+
+    # flavors selected always
+    my $start = '';
+    $start .= "-$_" for grep $self->{$_}, keys %flavor;
+
+    # make variations
+    my @var = $start;
+    for my $both (@both) {
+        my $append = join('-', $both, grep defined, $flavor{$both}{implies});
+        push @var, map "$_-$append", @var;
+    }
+
+    # normalize the variation names
+    @var = map { join '-', '', sort { $flavor{$a}{ix} <=> $flavor{$b}{ix} } grep length, split /-+/, $_ } @var;
+    s/(\b\w+\b)(?:-\1)+/$1/g for @var; # remove duplicate flavors
+
+    if ($Config::Config{archname64} eq '') {
+        # this is a 64bit platform. 64int and 64all are always set but
+        # we don't want them to appear on the final perl name
+        s/-64\w+//g for @var;
+    }
+
+    # remove duplicated variations
+    my %var = map { $_ => 1 } @var;
+    sort keys %var;
+}
+
+sub run_command_install_multiple {
+    my ( $self, @dists) = @_;
+
+    unless(@dists) {
+        $self->run_command_help("install-multiple");
+        exit(-1);
+    }
+
+    die "--switch can not be used with command install-multiple.\n\n"
+        if $self->{switch};
+    die "--as can not be used when more than one distribution is given.\n\n"
+        if $self->{as} and @dists > 1;
+
+    my @variations = $self->check_and_calculate_variations;
+    print join("\n",
+               "Compiling the following distributions:",
+               map("    $_", @dists),
+               "  with the following variations:",
+               map((/-(.*)/ ? "    $1" : "    default"), @variations),
+               "", "");
+
+    my @ok;
+    for my $dist (@dists) {
+        for my $variation (@variations) {
+            local $@;
+            eval {
+                $self->{$_} = '' for keys %flavor;
+                $self->{$_} = 1 for split /-/, $variation;
+                $self->{variation} = $variation;
+                $self->{installation_name} = undef;
+
+                $self->run_command_install($dist);
+                push @ok, $self->{installation_name};
+            };
+            if ($@) {
+                $@ =~ s/\n+$/\n/;
+                print "Installation of $dist$variation failed: $@";
+            }
+        }
+    }
+
+    print join("\n",
+               "",
+               "The following perls have been installed:",
+               map ("    $_", grep defined, @ok),
+               "", "");
+    return
+}
+
 sub run_command_download {
     my ($self, $dist) = @_;
 
@@ -1031,16 +1213,20 @@ sub do_install_archive {
 
 sub do_install_this {
     my ($self, $dist_extracted_dir, $dist_version, $installation_name) = @_;
-    $self->{dist_extracted_dir} = $dist_extracted_dir;
-    $self->{log_file} ||= joinpath($self->root, "build.${installation_name}.log");
 
-    my $version = perl_version_to_integer($dist_version);
+    my $variation = $self->{variation};
+
+    $self->{dist_extracted_dir} = $dist_extracted_dir;
+    $self->{log_file} = joinpath($self->root, "build.${installation_name}${variation}.log");
 
     my @d_options = @{ $self->{D} };
     my @u_options = @{ $self->{U} };
     my @a_options = @{ $self->{A} };
     my $sitecustomize = $self->{sitecustomize};
     $installation_name = $self->{as} if $self->{as};
+    $installation_name .= $variation;
+
+    $self->{installation_name} = $installation_name;
 
     if ( $sitecustomize ) {
         die "Could not read sitecustomize file '$sitecustomize'\n"
@@ -1048,6 +1234,14 @@ sub do_install_this {
         push @d_options, "usesitecustomize";
     }
 
+    if ( $self->{noman} ) {
+        push @d_options, qw/man1dir=none man3dir=none/;
+    }
+
+    for my $flavor (keys %flavor) {
+        $self->{$flavor} and push @d_options, $flavor{$flavor}{d_option}
+    }
+
     my $perlpath = $self->root . "/perls/$installation_name";
     my $patchperl = $self->root . "/bin/patchperl";
 
@@ -1062,7 +1256,8 @@ sub do_install_this {
         push @a_options, "'eval:scriptdir=${perlpath}/bin'";
     }
 
-    if ( $version < perl_version_to_integer( '5.6.0' ) ) {
+    my $version = perl_version_to_integer($dist_version);
+    if (defined $version and $version < perl_version_to_integer( '5.6.0' ) ) {
         # ancient perls do not support -A for Configure
         @a_options = ();
     }
@@ -1090,7 +1285,7 @@ INSTALL
                 ( map { qq{'-U$_'} } @u_options ),
                 ( map { qq{'-A$_'} } @a_options ),
             ),
-        $version < perl_version_to_integer( '5.8.9' )
+        (defined $version and $version < perl_version_to_integer( '5.8.9' ))
                 ? ("$^X -i -nle 'print unless /command-line/' makefile x2p/makefile")
                 : ()
     );
@@ -1175,6 +1370,20 @@ sub do_install_program_from_url {
 
     my $body = http_get($url) or die "\nERROR: Failed to retrieve $program_name executable.\n\n";
 
+    unless ($body =~ m{\A#!/}s) {
+        my $x = joinpath($ENV{TMPDIR} || "/tmp", "${program_name}.downloaded.$$");
+        my $message = "\nERROR: The downloaded $program_name program seem to be invalid. Please check if the following URL can be reached correctly\n\n\t$url\n\n...and try again latter.";
+
+        unless (-f $x) {
+            open my $OUT, ">", $x;
+            print $OUT $body;
+            close($OUT);
+            $message .= "\n\nThe previously downloaded file is saved at $x for manual inspection.\n\n";
+        }
+
+        die $message;
+    }
+
     if ($body_filter && ref($body_filter) eq "CODE") {
         $body = $body_filter->($body);
     }
@@ -1297,7 +1506,6 @@ sub perlbrew_env {
 
         if ($lib_name) {
             require local::lib;
-            no warnings 'uninitialized';
 
             if ($ENV{PERL_LOCAL_LIB_ROOT}
                 && $ENV{PERL_LOCAL_LIB_ROOT} =~ /^\Q$PERLBREW_HOME\E/
@@ -1311,8 +1519,10 @@ sub perlbrew_env {
             if (-d $base) {
                 delete $ENV{PERL_LOCAL_LIB_ROOT};
                 @ENV{keys %env} = values %env;
+                while (my ($k,$v) = each %ENV) {
+                    delete $ENV{$k} unless defined($v);
+                }
                 my %lib_env = local::lib->build_environment_vars_for($base, 0, 1);
-
                 $env{PERLBREW_PATH}    = joinpath($base, "bin") . ":" . $env{PERLBREW_PATH};
                 $env{PERLBREW_MANPATH} = joinpath($base, "man") . ":" . $env{PERLBREW_MANPATH};
                 $env{PERLBREW_LIB}  = $lib_name;
@@ -1627,7 +1837,7 @@ sub run_command_install_patchperl {
 
 sub run_command_install_cpanm {
     my ($self) = @_;
-    $self->do_install_program_from_url('https://github.com/miyagawa/cpanminus/raw/master/cpanm' => 'cpanm');
+    $self->do_install_program_from_url('https://raw.github.com/miyagawa/cpanminus/master/cpanm' => 'cpanm');
 }
 
 sub run_command_self_upgrade {
@@ -1740,10 +1950,16 @@ sub run_command_clean {
     my @build_dirs = <$root/build/*>;
 
     for my $dir (@build_dirs) {
-        print "Remove $dir\n";
+        print "Removing $dir\n";
         rmpath($dir);
     }
 
+    my @tarballs = <$root/dists/*>;
+    for my $file ( @tarballs ) {
+        print "Removing $file\n";
+        unlink($file);
+    }
+
     print "\nDone\n";
 }
 
@@ -1990,7 +2206,7 @@ sub run_command_upgrade_perl {
     print "Upgrading $current->{name} to $dist_version\n" unless $self->{quiet};
     local $self->{as}        = $current->{name};
     local $self->{dist_name} = $dist;
-    $self->do_install_release($dist, "perl", $dist_version);
+    $self->do_install_release($dist, $dist_version);
 }
 
 sub run_command_list_modules {
@@ -2479,7 +2695,7 @@ close to what your want to read.
 
 =head1 INSTALLATION
 
-It is the simpleist to use the perlbrew installer, just paste this statement to
+It is the simplest to use the perlbrew installer, just paste this statement to
 your terminal:
 
     curl -kL http://install.perlbrew.pl | bash
@@ -2493,7 +2709,7 @@ should follow the instruction on screen to modify your shell rc file to put it
 in your PATH.
 
 The installed perlbrew command is a standalone executable that can be run with
-system perl. The minimun system perl version requirement is 5.8.0, which should
+system perl. The minimum system perl version requirement is 5.8.0, which should
 be good enough for most of the OSes these days.
 
 A packed version of C<patchperl> to C<~/perl5/perlbrew/bin>, which is required
